!+
! Module: rgscelestoffset_mod
!
! Authors:  C. Erd -  Astrophysics/XMM-Newton SOC/ESA
!          
! Description: Calculate a celestial offset position based on input in RGS FOV coordinates.
! 
! Comments: Note: CAL and OAL musr be properly initialized before calling this routine!
! 
! $Id: rgscelestoffset_mod.f90,v 1.2 2000/08/11 13:08:45 pvideler Exp $
!- 
module rgscelestoffset_mod

!PV  use dal
!PV  use cal
!PV  use param  
  use oal
  use errorHandling
  use rgslib

 ! ......................................................
  implicit none 
  private 
  public rgscelestoffset
 ! ......................................................

  ! global variables (by CE)
  type(SpacecraftAttitudeType) :: ref !instrument reference attitude
  real (double), dimension(2)  :: offaxis !offset in RGS FOV coordinates

contains 

    !---------------------------------------------------------------------
!  function rgscelestoffset(att0, offset)   result(srcpos)
function rgscelestoffset(att0, disp, xdisp)   result(srcpos)
    !---------------------------------------------------------------------
    ! new wrapper function by PV. 11-aug-2000
    ! Note: CAL and OAL musr be properly initialized before calling this routine!
!*** 15 Mar 2001 - changed vector argument to 2 scalars - IMS.

    type(SpacecraftAttitudeType), intent(in) :: att0 ! reference attitude (rad)
!    real(double), dimension(2), intent(in) :: offset ! (rad)
real(double), intent(in) :: disp, xdisp ! (rad)
real(double), dimension(2) :: offset ! (rad)
    type(SkyPositionType)                  :: srcpos ! (rad) 

    real(double), dimension(2) :: src

offset(1) = disp
offset(2) = xdisp

    ref = att0           ! reference attitude in radians
    offaxis = offset     ! offaxis angles in radians

    src(1) = ref%ra             !starting values
    src(2) = ref%dec

    call minimize(src)

    srcpos%ra  = src(1)
    srcpos%dec = src(2)

  end function rgscelestoffset

  
!!$  !---------------------------------------------------------------------
!!$  subroutine rgscelestoffset
!!$  !---------------------------------------------------------------------
!!$    real(double), dimension(2) :: src
!!$
!!$    ref%ra   = realParameter("ra")   /180d0*pi !reference attitude
!!$    ref%dec  = realParameter("dec")  /180d0*pi
!!$    ref%apos = realParameter("apos") /180d0*pi
!!$
!!$    offaxis(1) = realParameter("disp")/36d2/180d0*pi !offset in dispersion in arcsec
!!$    offaxis(2) = realParameter("xdsp")/36d2/180d0*pi !offset in cross-dispersion in arcsec
!!$
!!$    call CAL_setState(instrument=instId(stringParameter("instrument")))
!!$    src(1) = ref%ra             !starting values
!!$    src(2) = ref%dec
!!$
!!$    call minimize( src )
!!$
!!$    write (*,*) 'Offaxis source is at Ra,Dec=',src*180d0/pi    
!!$  end subroutine rgscelestoffset

  
  !---------------------------------------------------------------------
  subroutine minimize (srccoord)          !interface to SLATEC library
  !---------------------------------------------------------------------
    real(double), dimension(2) :: srccoord !this is the position that should be fit

                                !for interface to DNLS1E
    integer, parameter ::   n = 2 !number of variables
    INTEGER IOPT,M,NPRINT,INFO,LWA,IW(N)
    real(double) :: TOL,X(N)
    real(double), dimension(:), allocatable :: FVEC, WA !FVEC(M),WA(LWA)

                                !for interface to DCOV
    real(double), dimension(:,:), allocatable :: R !holds covariance matrix
    integer :: LDR
    real(double), dimension(:), allocatable  :: WA1, WA2, WA3, WA4

    real(double) :: D1MACH      !machine precision
    integer :: i

    iopt = 1                    !do not calculate the Jakobian in FCN
    m = 2                       !number of functions
    lwa = N*(M+5)+M
    allocate(FVEC(m), WA(lwa))
    TOL = SQRT(D1MACH(4))/1000.       !distance to solution
    nprint = 20                  !if >0 enable printing in FCN every NPRINT times
    iw = n
    wa = 0
    x = srccoord
    fvec = 0
    write (errstr,*) 'calling DNLS1E now'
    call message (appMsg, Sparse)
    call DNLS1E (FCN, IOPT, M, N, X, FVEC, TOL, NPRINT, INFO, IW, WA, LWA)
    select case (info)
    case (0)
       write (errstr,*) 'DNLS1E(INFO): improper input parameters.'
    case (1)
       write (errstr,*) 'DNLS1E(INFO): algorithm estimates that the relative error in the sum of squares is at most TOL.'
    case (2)
       write (errstr,*) 'DNLS1E(INFO): algorithm estimates that the relative error between X and the solution is at most TOL.'
    case (3)
       write (errstr,*) 'DNLS1E(INFO): conditions for INFO = 1 and INFO = 2 both hold.'
    case (4)
       write (errstr,*) 'DNLS1E(INFO): FVEC is orthogonal to the columns of the Jacobian to machine precision.'
    case (5)
       write (errstr,*) 'DNLS1E(INFO): number of calls to FCN has reached 100*(N+1) for IOPT=2 or 3 or 200*(N+1) for IOPT=1.'
    case (6)
       write (errstr,*) 'DNLS1E(INFO): TOL is too small.  No further reduction in the sum of squares is possible.'
    case (7)
       write (errstr,*) 'DNLS1E(INFO): TOL is too small.  No further improvement in the approximate solution X is possible.'
    case default
       write (errstr,*) 'DNLS1E(INFO): successful completion.'
    end select
    call message (appMsg, Verbose)
    srccoord = x

    allocate(R(m,n), WA1(N), WA2(N), WA3(N), WA4(M))
    r = 0
    LDR = M
    call DCOV (FCN, IOPT, M, N, X, FVEC, R, LDR, INFO, WA1, WA2, WA3, WA4)
    select case (info)
    case (0)
       write (errstr,*) 'DCOV(INFO): Improper input parameters (M.LE.0 or N.LE.0).'
    case (1)
       write (errstr,*) 'DCOV(INFO): Successful return. The covariance matrix has been calculated' &
            ,' and stored in the upper N by N submatrix of R.'
    case (2)
       write (errstr,*) 'DCOV(INFO): The Jacobian matrix is singular for the input value of X.' &
            ,' The covariance matrix cannot be calculated. The upper N by N submatrix of R' &
            ,' contains the QR factorization of the Jacobian (probably not of interest to' &
            ,' the user).'
    case default
       write (errstr,*) 'DCOV(INFO): INFO =', info
    end select
    call message (appMsg, Verbose)

    write (errstr,*) 'Covariance matrix:'
    call message (appMsg, Noisy)
    do i=1,n
       write (errstr,*) r(1:n,i)
       call message (appMsg, Noisy)
    end do

    deallocate(FVEC, WA, R, WA1, WA2, WA3, WA4)
  end subroutine minimize

  !---------------------------------------------------------------------
  SUBROUTINE FCN(IFLAG,M,N,X,FVEC,FJAC,LDFJAC)
  !---------------------------------------------------------------------
                                !have to specify the arguments as F77 expects
    INTEGER IFLAG,LDFJAC,M,N
    real(double) :: X(N),FVEC(M),FJAC(LDFJAC,N)
real(double) :: dummy(LDFJAC,N)

    type(skyPositionType)                  :: src

       ! If IFLAG=0, the values in X and FVEC are available
       ! for printing.
       ! If IFLAG=1, calculate the functions at X and return
       ! this vector in FVEC.
       ! If IFLAG=2, calculate the full Jacobian at X and return
       ! this matrix in FJAC.  Note that IFLAG will never be 2 unless
       ! IOPT=2.  FVEC contains the function values at X and must
       ! not be altered.  FJAC(I,J) must be set to the derivative
       ! of FVEC(I) with respect to X(J).
       ! If IFLAG=3, calculate the LDFJAC-th row of the Jacobian
       ! and return this vector in FJAC.  Note that IFLAG will
       ! never be 3 unless IOPT=3.  FVEC contains the function
       ! values at X and must not be altered.  FJAC(J) must be
       ! set to the derivative of FVEC(LDFJAC) with respect to X(J).

dummy(1,1) = FJAC(1,1) ! just to stop the compiler warnings.

    if (iflag == 0) then
       write (errstr,*) 'FCN: X=',x*180d0/pi*36d2,'FVEC=',FVEC
       call message (appMsg, Noisy)
    endif

    if (iflag == 1) then
       src%ra = x(1)
       src%dec = x(2)
       fvec = offAxisAngles( ref, src ) / (/RGSLIB_getFoverL(),1d0/) - offaxis
    end if
  end SUBROUTINE FCN
end module rgscelestoffset_mod






